@@ -52,6 +52,27 @@ def do_list_filters(cs, args):
         shell_utils.print_resource_filter_list(filters)
 
 
+@utils.arg('--filters',
+           type=six.text_type,
+           nargs='*',
+           start_version='3.52',
+           metavar='<key=value>',
+           default=None,
+           help="Filter key and value pairs. Admin only.")
+def do_type_list(cs, args):
+    """Lists available 'volume types'.
+
+    (Only admin and tenant users will see private types)
+    """
+    # pylint: disable=function-redefined
+    search_opts = {}
+    # Update search option with `filters`
+    if hasattr(args, 'filters') and args.filters is not None:
+        search_opts.update(shell_utils.extract_filters(args.filters))
+    vtypes = cs.volume_types.list(search_opts=search_opts)
+    shell_utils.print_volume_type_list(vtypes)
+
+
 @utils.arg('--all-tenants',
            metavar='<all_tenants>',
            nargs='?',
@@ -484,10 +505,6 @@ def do_reset_state(cs, args):
            help='Creates volume from volume ID. Default=None.')
 @utils.arg('--source_volid',
            help=argparse.SUPPRESS)
-@utils.arg('--source-replica',
-           metavar='<source-replica>',
-           default=None,
-           help='Creates volume from replicated volume ID. Default=None.')
 @utils.arg('--image-id',
            metavar='<image-id>',
            default=None,
@@ -547,7 +564,7 @@ def do_reset_state(cs, args):
 @utils.arg('--allow-multiattach',
            dest='multiattach',
            action="store_true",
-           help=('Allow volume to be attached more than once.'
+           help=('Allow volume to be attached more than once. (DEPRECATED)'
                  ' Default=False'),
            default=False)
 @utils.arg('--poll',
@@ -604,7 +621,6 @@ def do_create(cs, args):
                                imageRef=image_ref,
                                metadata=volume_metadata,
                                scheduler_hints=hints,
-                               source_replica=args.source_replica,
                                multiattach=args.multiattach,
                                backup_id=backup_id)
 
@@ -1080,12 +1096,19 @@ def do_cluster_disable(cs, args):
 @utils.arg('--resource-type', metavar='<Volume|Snapshot>', default=None,
            choices=('Volume', 'Snapshot'),
            help='Type of resource to cleanup.')
+@utils.arg('--service-id',
+           metavar='<service-id>',
+           type=int,
+           default=None,
+           help='The service id field from the DB, not the uuid of the'
+                ' service. Default=None.')
 def do_work_cleanup(cs, args):
     """Request cleanup of services with optional filtering."""
     filters = dict(cluster_name=args.cluster, host=args.host,
                    binary=args.binary, is_up=args.is_up,
                    disabled=args.disabled, resource_id=args.resource_id,
-                   resource_type=args.resource_type)
+                   resource_type=args.resource_type,
+                   service_id=args.service_id)
 
     filters = {k: v for k, v in filters.items() if v is not None}
 
@@ -1516,6 +1539,7 @@ def do_group_list_replication_targets(cs, args):
         utils.print_list(rep_targets, [key for key in rep_targets[0].keys()])
 
 
+@api_versions.wraps('3.14')
 @utils.arg('--all-tenants',
            dest='all_tenants',
            metavar='<0|1>',
@@ -2069,6 +2093,8 @@ def do_attachment_list(cs, args):
                                       marker=args.marker,
                                       limit=args.limit,
                                       sort=args.sort)
+    for attachment in attachments:
+        setattr(attachment, 'server_id', getattr(attachment, 'instance', None))
     columns = ['ID', 'Volume ID', 'Status', 'Server ID']
     if args.sort:
         sortby_index = None
@@ -2135,6 +2161,13 @@ def do_attachment_show(cs, args):
            metavar='<mountpoint>',
            default=None,
            help='Mountpoint volume will be attached at. Default=None.')
+@utils.arg('--mode',
+           metavar='<mode>',
+           default='null',
+           start_version='3.54',
+           help='Mode of attachment, rw, ro and null, where null '
+                'indicates we want to honor any existing '
+                'admin-metadata settings.  Default=null.')
 def do_attachment_create(cs, args):
     """Create an attachment for a cinder volume."""
 
@@ -2149,9 +2182,12 @@ def do_attachment_create(cs, args):
                      'multipath': args.multipath,
                      'mountpoint': args.mountpoint}
     volume = utils.find_volume(cs, args.volume)
+    mode = getattr(args, 'mode', 'null')
     attachment = cs.attachments.create(volume.id,
                                        connector,
-                                       args.server_id)
+                                       args.server_id,
+                                       mode)
+
     connector_dict = attachment.pop('connection_info', None)
     utils.print_dict(attachment)
     if connector_dict:
@@ -2331,6 +2367,11 @@ def do_service_get_log(cs, args):
            default=None,
            start_version='3.43',
            help='Metadata key and value pairs. Default=None.')
+@utils.arg('--availability-zone',
+           default=None,
+           start_version='3.51',
+           help='AZ where the backup should be stored, by default it will be '
+           'the same as the source.')
 def do_backup_create(cs, args):
     """Creates a volume backup."""
     if args.display_name is not None:
@@ -2339,25 +2380,22 @@ def do_backup_create(cs, args):
     if args.display_description is not None:
         args.description = args.display_description
 
-    volume = utils.find_volume(cs, args.volume)
-    if hasattr(args, 'metadata') and args.metadata:
-        backup = cs.backups.create(volume.id,
-                                   args.container,
-                                   args.name,
-                                   args.description,
-                                   args.incremental,
-                                   args.force,
-                                   args.snapshot_id,
-                                   shell_utils.extract_metadata(args))
-    else:
-        backup = cs.backups.create(volume.id,
-                                   args.container,
-                                   args.name,
-                                   args.description,
-                                   args.incremental,
-                                   args.force,
-                                   args.snapshot_id)
+    kwargs = {}
+    if getattr(args, 'metadata', None):
+        kwargs['metadata'] = shell_utils.extract_metadata(args)
+    az = getattr(args, 'availability_zone', None)
+    if az:
+        kwargs['availability_zone'] = az
 
+    volume = utils.find_volume(cs, args.volume)
+    backup = cs.backups.create(volume.id,
+                               args.container,
+                               args.name,
+                               args.description,
+                               args.incremental,
+                               args.force,
+                               args.snapshot_id,
+                               **kwargs)
     info = {"volume_id": volume.id}
     info.update(backup._info)
 
@@ -2365,3 +2403,38 @@ def do_backup_create(cs, args):
         info.pop('links')
 
     utils.print_dict(info)
+
+
+@utils.arg('volume', metavar='<volume>',
+           help='Name or ID of volume to transfer.')
+@utils.arg('--name',
+           metavar='<name>',
+           default=None,
+           help='Transfer name. Default=None.')
+@utils.arg('--display-name',
+           help=argparse.SUPPRESS)
+@utils.arg('--no-snapshots',
+           action='store_true',
+           help='Allows or disallows transfer volumes without snapshots. '
+                'Default=False.',
+           start_version='3.55',
+           default=False)
+def do_transfer_create(cs, args):
+    """Creates a volume transfer."""
+    if args.display_name is not None:
+        args.name = args.display_name
+
+    kwargs = {}
+    no_snapshots = getattr(args, 'no_snapshots', None)
+    if no_snapshots is not None:
+        kwargs['no_snapshots'] = no_snapshots
+
+    volume = utils.find_volume(cs, args.volume)
+    transfer = cs.transfers.create(volume.id,
+                                   args.name,
+                                   **kwargs)
+    info = dict()
+    info.update(transfer._info)
+
+    info.pop('links', None)
+    utils.print_dict(info)
